from .errors import MakeResponseError


class Response(object):

    def __init__(self, request_id, code, data, message):
        self.code = code
        self.data = data
        self.message = message
        self.request_id = request_id
    
    @classmethod
    def load_from_dict(cls, data):
        if not isinstance(data, dict):
            raise MakeResponseError("Invalid data type, dict is expected, but {} found".format(type(data)))
        if not data.get("request_id"):
            raise MakeResponseError("request_id is necessary for response")
        if data.get("code", None) is None:
            raise MakeResponseError("code is necessary for response")
        return cls(data["request_id"], data["code"], data.get("data"), data.get("message"))
    
    def json(self):
        return self.__dict__





class ValidateError(Exception):
    """
    参数验证失败
    """
    def __init__(self, message="validate params failed"):
        Exception.__init__(self, message)


class HandlerNotFound(Exception):
    """
    未找到 handler
    """
    def __init__(self, message="handler not found"):
        Exception.__init__(self, message)


class MakeHandlerResponseError(Exception):
    """
    handler 返回的数据不合法
    """
    def __init__(self, message="response type error"):
        Exception.__init__(self, message)


class InvalidHandler(Exception):
    """
    非法的 handler，handler 必须继承于 Handler 类
    """
    def __init__(self, message="invalid handler"):
        Exception.__init__(self, message)


class MakeCommandError(Exception):
    """
    创建 Command 失败
    """
    def __init__(self, message="failed to init command"):
        Exception.__init__(self, message)


class InvalidMessage(Exception):
    """
    未能识别的消息，消息必须能转化成 Command
    """
    def __init__(self, message="invalid message"):
        Exception.__init__(self, message)


class MakeResponseError(Exception):
    """
    创建 Response 失败
    """
    def __init__(self, message="make response error"):
        Exception.__init__(self, message)


class MakeRequestError(Exception):
    """
    创建 Request 失败
    """
    def __init__(self, message="make response error"):
        Exception.__init__(self, message)



import uuid
import json
import redis

from .param import Param
from .command import Command
from .response import Response
from .handler import Handler, HandlerOK, HandlerFailed
from .errors import InvalidHandler, HandlerNotFound, InvalidMessage, MakeRequestError


class Laka(object):
    """
    Laka is a microservice framework based on json and redis
    """

    def __init__(self, redis_host, redis_port, redis_queue, response_message=None, redis_db=0):
        self.redis_host = redis_host
        self.redis_port = redis_port
        self.redis_db = redis_db
        self.request_queue = redis_queue
        self.redis_client = None
        if response_message and not isinstance(response_message, dict):
            raise TypeError("Invalid type of response_message, dict is expected but {} found".format(type(response_message)))
        self._connect_redis()
        self.response_message = response_message if response_message else {}
        self.handlers = {}

    def _connect_redis(self):
        self.redis_client = redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db)
    
    def register(self, command_code, handler):
        if not issubclass(handler, Handler):
            raise InvalidHandler("handler should be subclass of Handler, but {} found".format(handler))
        if handler.Param and not issubclass(handler.Param, Param):
            raise InvalidHandler("handler.Param should be subclass of Param, but {} found".format(handler.Param))
        self.handlers[command_code] = handler

    def handle(self, cmd):
        handler = self.handlers.get(cmd.code)
        if not handler:
            raise HandlerNotFound("handler not found for CommandCode = {}".format(cmd.code))
        h = handler()
        h.get_param(cmd)
        return h.handle()

    def request(self, code, param):
        if not isinstance(param, Param):
            raise MakeRequestError("param should an object of Param")
        request_id = self.new_request_id()
        cmd = Command(code, param, request_id)
        try:
            r = json.dumps(cmd.json())
        except Exception as e:
            raise MakeRequestError(e)
        self.redis_client.lpush(self.request_queue, r)
        return request_id

    def _accept(self, queue):
        d = self.redis_client.brpop(queue)
        try:
            data = json.loads(str(d[1], encoding="utf8"))
        except Exception as e:
            raise InvalidMessage(e)
            return None
        return data
    
    def accept_request(self):
        while True:
            data = self._accept(self.request_queue)
            if not data:
                continue
            cmd = Command.load_from_dict(data)
            yield cmd
        
    def accept_response(self, request_id):
        data = self._accept(request_id)
        response = Response.load_from_dict(data)
        return response

    def response(self, request_id, handler_response):
        """
        由于发送返回数据时，需要知道 request_id，因此如果请求时数据格式不正确或者缺少 request_id，则该请求将得不到任何返回值。
        所以必须确保发送请求时数据格式正确。
        """
        if not isinstance(handler_response, (HandlerFailed, HandlerOK)):
            raise MakeResponseError("the type of handler_response should be HandlerOK or HandlerFailed, but {} found".format(type(handler_response)))
        message = self.response_message.get(handler_response.code)
        resp = Response(request_id, handler_response.code, handler_response.data, message)
        try:
            data = json.dumps(resp.json())
        except Exception as e:
            raise MakeResponseError(e)
        self.redis_client.lpush(request_id, data)

    def new_request_id(self):
        return "LAKA:REQUEST_ID:{}".format(uuid.uuid4().hex)


from .errors import MakeHandlerResponseError


class Handler(object):
    Param = None

    def __init__(self):
        self.param = None

    def get_param(self, cmd):
        if not self.Param:
            return
        self.param = self.Param.load_from_cmd(cmd)


class HandlerResponse(object):
    pass


class HandlerFailed(HandlerResponse):

    def __init__(self, code):
        if code is None:
            raise MakeHandlerResponseError("response code should not be None")
        self.code = code
        self.data = None


class HandlerOK(HandlerResponse):
    code = None

    def __init__(self, data):
        if self.code is None:
            raise MakeHandlerResponseError("you should call function HandlerOK.set_success_code to set success code")
        if data and not isinstance(data, dict):
            raise MakeHandlerResponseError("response type error, dict is expected but {} found".format(type(data)))
        self.data = data
    
    @classmethod
    def set_success_code(cls, code):
        cls.code = code


from .errors import ValidateError

class Param(object):
    
    def __init__(self):
        raise NotImplementedError
    
    def validate(self):
        return True
    
    @classmethod
    def load_from_cmd(cls, cmd):
        param = cls()
        keys = param.__dict__.keys()
        for key in keys:
            setattr(param, key, cmd.params.get(key))
        status = param.validate()
        if not status:
            raise ValidateError("Invalid params.")
        return param

    def json(self):
        status = self.validate()
        if not status:
            raise ValidateError("Invalid params")
        return self.__dict__